
# Local descriptive exceptions
class NameTypeMismatch(Exception): pass
class BadSpecifiedType(Exception): pass
class BadSpecifiedName(Exception): pass
class NoNamesSpecified(Exception): pass
class ImproperUsage(Exception): pass
class MissingModule(Exception): pass
class UnknownName(Exception): pass
class Unsupported(Exception): pass
class BadAssignment(Exception): pass
class BadElement(Exception): pass
class BadTarget(Exception): pass
class BadValue(Exception): pass
class BadIndex(Exception): pass
class BadData(Exception): pass
class Empty(Exception): pass
